package stockMarketApplet.market.agents.technicalTrader.basics;

import java.util.List;
import java.util.Random;
import java.util.Calendar;
import java.util.ArrayList;
import stockMarketApplet.market.agents.technicalTrader.TechnicalTrader;
import stockMarketApplet.market.agents.technicalTrader.basics.Program;
import stockMarketApplet.market.io.DataRecord;

// Class that represents individuals of the GP mechanism
public class Individual extends Object implements Cloneable {
  private Program program;
  private double fitness= 0.0;
  private byte result;
  private Random generator;
  private long time;
  // Constructor method with a Program as parameter
  public Individual(Program prog)  {
    program = (Program) prog.clone();
    generator = new Random();
    Calendar date = Calendar.getInstance();
    generator.setSeed(date.get(Calendar.MILLISECOND));
  }
  // Method that implements the evaluation of the individual
  public final int evaluation(DataRecord dataRow)  {
    short tempShort = 0;
    return program.evaluation(tempShort,dataRow);
  }
  // Returns the root node of this individual
  public final Program getProgram()  {
    return program;
  }
  // Returns the fitness of this individual
  public final double getFitness()  {
    return fitness;
  }
  // Returns the value result of this individual's valuation
  public final byte getResult()  {
    return result;
  }
  // Returns the valuation time of this individual
  public long getTime()  {
    return time;
  }
  // Sets the fitness value of this individual
  public final void setFitness(double val)  {
    fitness = val;
  }
  // Sets the execution time of this individual
  public void setTime(long t)  {
    time = t;
  }
  // Sums a double value to the fitness value
  public final void addFitness(double d)  {
    fitness += d;
  }
  // Resets the fitness value
  public void resetFitness()  {
    fitness = 0.0;
  }
  // Returns the size of the node of this individual
  public int size()  {
    return program.size();
  }
  // Returns the depth of this individual's node
  public short depth()  {
    return program.depth();
  }
  // Returns the subtree under the node n of the individual's root node
  private Program subTree(short n)  {
    return new Program(program.subTree(n));
  }
  // Replaces the subtree of individual's root node under the node n
  private void replaceTree(short n, Program tree)  {
    if(n == 0)
      program = (Program) tree.clone();
    else
      program.replaceTree((short)n, tree.getExpression());
  }
  // Performs the crossover operation with the provided mate individual
  public Individual[] crossover(Individual mate)  {
    Individual[] offspring = new Individual[2];
    offspring[0] = (Individual) mate.clone();
    offspring[1] = (Individual) this.clone();
    short r1, r2;
    Program sub1, sub2;
    do {
      r1 = (short) (offspring[0].size() * generator.nextDouble());
      sub1 = offspring[0].subTree(r1);
      r2 = (short) (offspring[1].size() * generator.nextDouble());
      sub2 = offspring[1].subTree(r2);
    } while (!isCompatible(sub1, sub2));
    offspring[0].replaceTree(r1, sub2);
    offspring[1].replaceTree(r2, sub1);
    return offspring;
  }
  // Verifies if the two provided nodes are compatible, this is to asure that the trees remain sintactically correct
  private boolean isCompatible(Program n1, Program n2){
    return this.getProgram().isCompatible(n1.getExpression(),n2.getExpression());
  }
  // Mutation method that requires some parameters to generate new traders during the mutation process
  public Individual mutation( byte iniTreeDepth, double minThreshold, double rangeThreshold, List ind, List logicFunct,
                              List relFunct){
    Individual temp =  (Individual) this.clone();
    TechnicalTrader gedefit = new TechnicalTrader(ind, logicFunct, relFunct);
    List tempList = new ArrayList();
    gedefit.setIniTreeDepth(iniTreeDepth);
    gedefit.setMinThreshold(minThreshold);
    gedefit.setRangeThreshold(rangeThreshold);
    short r1 = (short) (temp.size() * generator.nextDouble());
    Program sub1,sub2;
    sub1 = temp.subTree(r1);
    if (  sub1.getExpression().get(0).toString().compareTo("ITE")==0 ||
          (sub1.getDecisionSet().contains(sub1.getExpression().get(0)))  )
      sub2 = new Program(gedefit.createRandTree(gedefit.iniTreeDepth));
    else if (sub1.isThreshold(sub1.getExpression().get(0).toString())){
      tempList.add(gedefit.createRandThreshold());
      sub2 = new Program(tempList);
    }
    else if (sub1.getIndicatorSet().contains(sub1.getExpression().get(0))){
      tempList.add(gedefit.createRandIndicator());
      sub2 = new Program(tempList);
    }
    else
      sub2 = new Program(gedefit.createRandCond(gedefit.iniTreeDepth));
    temp.replaceTree(r1, sub2);
    return temp;
  }
  // Maximization method to optimize the threshold values
  public Individual maximization( short nodeNum, Program subtree, double r, int d, DataRecord[] data, int[] recommendations,
                                  double w_rc, double w_rf, double w_rmc){
    Individual temp = (Individual) this.clone();
    String tempStr = new String();
    short nodeNumber = nodeNum;
    short tempShort1 = (short)(nodeNumber+1);
    short tempShort2;
    short tempShort3;
    if (subtree.depth()<2){
      return this;
    }
    else if (subtree.depth() == 2){
      return adjustThreshold(nodeNumber, temp, r, d, data, recommendations,  w_rc,  w_rf,  w_rmc);
    }
    else {
      tempStr = (String)this.getProgram().getExpression().get(nodeNumber);
      if (tempStr.compareTo("ITE")==0){
        temp=temp.maximization(tempShort1,new Program(this.getProgram().subTree(tempShort1)),r,d,data,recommendations,w_rc,w_rf,w_rmc);
        tempShort2=(short)(this.getProgram().subTree(tempShort1).size()+tempShort1);
        temp=temp.maximization(tempShort2,new Program(this.getProgram().subTree(tempShort2)),r,d,data,recommendations,w_rc,w_rf,w_rmc);
        tempShort3=(short)(this.getProgram().subTree(tempShort2).size()+tempShort2);
        temp=temp.maximization(tempShort3,new Program(this.getProgram().subTree(tempShort3)),r,d,data,recommendations,w_rc,w_rf,w_rmc);
      }
      else if (tempStr.compareTo("NOT")==0){
        temp=temp.maximization(tempShort1,new Program(this.getProgram().subTree(tempShort1)),r,d,data,recommendations,w_rc,w_rf,w_rmc);
      }
      else{
        temp=temp.maximization(tempShort1,new Program(this.getProgram().subTree(tempShort1)),r,d,data,recommendations,w_rc,w_rf,w_rmc);
        tempShort2=(short)(this.getProgram().subTree(tempShort1).size()+tempShort1);
        temp=temp.maximization(tempShort2,new Program(this.getProgram().subTree(tempShort2)),r,d,data,recommendations,w_rc,w_rf,w_rmc);
      }
      return temp;
    }
  }
  // Auxiliar method to the maximization operator
  private Individual adjustThreshold( short n, Individual ind, double rate, int days, DataRecord[] data, int[] recommendations,
                                      double w_rc, double w_rf, double w_rmc){
    boolean local = true;
    double step = 7.0;
    Individual temp1 = new Individual((Program)ind.getProgram().clone());
    Individual temp2 = new Individual((Program)ind.getProgram().clone());
    Individual bestSoFar = new Individual((Program)ind.subTree((short)0).clone());
    double dControl=0;
    try{
      dControl = Double.parseDouble(bestSoFar.getProgram().subTree((short)(n+2)).get(0).toString());
    }
    catch(NumberFormatException e){
      System.out.println(this.toString());
      System.out.println(bestSoFar.getProgram().subTree((short)(n+2)).get(0).toString());
      System.out.println("Invalid Double Number");
    }
    TechnicalTrader trader = new TechnicalTrader(rate, days, recommendations,  w_rc,  w_rf,  w_rmc);
    bestSoFar = trader.evalInd(bestSoFar, data);
    do{
      temp1.getProgram().deleteElemExpression(n+2);
      temp1.getProgram().insertElemExpresion(n+2,String.valueOf(dControl+step));
      temp2.getProgram().deleteElemExpression(n+2);
      temp2.getProgram().insertElemExpresion(n+2,String.valueOf(dControl-step));
      temp1 = trader.evalInd(temp1, data);
      temp2 = trader.evalInd(temp2, data);
      if (temp1.getFitness() >= temp2.getFitness()){
        if ( temp1.getFitness() > bestSoFar.getFitness() ){
          dControl += step;
          bestSoFar.getProgram().deleteElemExpression(n+2);
          bestSoFar.getProgram().insertElemExpresion(n+2,String.valueOf(dControl));
          bestSoFar = trader.evalInd(bestSoFar, data);
        }
        else{
          local = false;
        }
      }
      else if (temp2.getFitness() > bestSoFar.getFitness()){
          dControl -= step;
          bestSoFar.getProgram().deleteElemExpression(n+2);
          bestSoFar.getProgram().insertElemExpresion(n+2,String.valueOf(dControl));
          bestSoFar = trader.evalInd(bestSoFar, data);
      }
      else{
        local = false;
      }
    }while(local);
    return bestSoFar;
  }
  public void setExpression(List expr){
    this.program.setExpression(expr);
  }
  // Method that clones this individual
  public Object clone()  {
    Individual ind = new Individual((Program) program.clone());
    ind.setFitness(this.getFitness());
    return (Object) ind;
  }
  // Method that prints the current individual
  public void print()  {
    System.out.println("depth: " + program.depth() + " size: " + program.size());
    System.out.print(" program: ");
    program.print();
    System.out.println("");
    System.out.println("fitness: " + fitness );
  }
  // Method that converts the current individual into its corresponding string representation
  public String toString()  {
    return ( /*"Individual depth: " + program.depth() + " size: " + program.size() +
      " fitness: " +*/ fitness +/* "\n" +*/ program.toString() );
  }
}